#ifndef MAMMUTPARSER_H
#define MAMMUTPARSER_H

#include <vector>
#include "Debug.h"

#define YYEMPTY		-2
#define YYEOF		0
#define YYTERROR	1

#ifndef YYINITDEPTH
#define YYINITDEPTH 200
#endif

// TODO: document the special reqirements on the FlexClass (FlexClass::ParserTypes, std::iostrem etc) (concepts)
namespace Mammut
{
	// Tokenizers need to support the method NextToken(...)
	template<typename FlexClass>
	class Flex
	{
	public:
		Flex(std::istream & is) { m_lex.yyin = &is; }
		int NextToken(typename FlexClass::ParserTypes::STYPE & lval, typename FlexClass::ParserTypes::LTYPE)
		{
			return m_lex.yylex(lval);
		}

	private:
		FlexClass m_lex;
	};

	template<typename FlexClass>
	class FlexWrapper
	{
	public:
		FlexWrapper(FlexClass & flexer) : m_lex(flexer) {}
		int NextToken(typename FlexClass::ParserTypes::STYPE & lval, typename FlexClass::ParserTypes::LTYPE)
		{
			return m_lex.yylex(lval);
		}

	private:
		FlexClass & m_lex;
	};

	template<typename FlexClass> FlexWrapper<FlexClass> flexWrapper(FlexClass & flexer)
	{
		return FlexWrapper<FlexClass>(flexer);
	}

	template<typename Lexer, typename Grammar> int parse(Lexer & lexer, Grammar & grammar)
	{
		return parse<NoDebug>(lexer, grammar);
	}

	template<typename Debug, typename Lexer, typename Grammar>
		int parse(Lexer & lexer, Grammar & grammar)
	{
		enum State
		{
			yynewstate,
			yybackup,
			yydefault,
			yyreduce,
			yyerrlab,	// here on detecting error
			yyerrpop,	// pop the current state because it cannot handle the error token
			yyerrhandle
		};

		typename Grammar::STYPE yylval;
		typename Grammar::LTYPE yylloc;

		yylloc.first_line = 0;
		yylloc.last_line = 0;
		yylloc.first_column = 0;
		yylloc.last_column = 0;
		yylloc.text = 0;

		int yystate = 0;
		int yyn;
		int yyerrstatus = 0; // number of tokens to shift before error messages enabled
		int yychar1=0; //  lookahead token as an internal (translated) token number
		int yychar = YYEMPTY; // Cause a token to be read.

		typedef std::vector<int> StateStack;
		StateStack stateStack;

		typename Grammar::STYPE yyval; //  the variable used to return semantic values from actions
		typedef std::vector<typename Grammar::STYPE> ValueStack;
		ValueStack valStack;
		valStack.resize(1);

		//typedef std::vector<Grammar::LTYPE> LocationStack;
		//LocationStack locStack;
		//locStack.resize(1);
		//locStack.reserve(YYINITDEPTH);

		Debug::Message("Starting parse");

		// Push a new state, which is found in  yystate.
		// In all cases, when you get here, the value and location stacks
		// have just been pushed. so pushing a state here evens the stacks.

		State parseState=yynewstate;
		for(;;)
		{
			switch(parseState)
			{
			case yynewstate:
				{
					stateStack.push_back(yystate);
					Debug::EnterState(yystate);
					parseState=yybackup;
					continue;
				}
			case yybackup:
				{
					// Do appropriate processing given the current state.
					// Read a lookahead token if we need one and don't already have one.
					// First try to decide what to do without reference to lookahead token.

					yyn = Grammar::yypact[yystate];
					if (yyn == Grammar::YYFLAG)
					{
						parseState=yydefault;
						continue;
					}

					// Not known => get a lookahead token if don't already have one.

					// yychar is either YYEMPTY or YYEOF
					// or a valid token in external form.
					if (yychar == YYEMPTY)
					{
						Debug::Message("Reading a token:");
						yychar = lexer.NextToken(yylval, yylloc);
					}

					// Convert token to internal form (in yychar1) for indexing tables with
					if (yychar <= 0) // This means end of input.
					{
						yychar1 = 0;
						yychar = YYEOF; // Don't call YYLEX any more
						Debug::Message("Now at end of input.");
					}
					else
					{
						yychar1 = Grammar::Translate(yychar);
						Debug::WriteNextToken(yychar, Grammar::yytname[yychar1]);
					}

					yyn += yychar1;
					if (yyn < 0 || yyn > Grammar::YYLAST || Grammar::yycheck[yyn] != yychar1)
					{
						parseState=yydefault;
						continue;
					}

					yyn = Grammar::yytable[yyn];

					//yyn is what to do for this token type in this state.
					//Negative => reduce, -yyn is rule number.
					//Positive => shift, yyn is new state.
					//New state is final state => don't bother to shift,
					//just return success.
					//0, or most negative number => error.

					if (yyn < 0)
					{
						if (yyn == Grammar::YYFLAG)
						{
							parseState=yyerrlab;
							continue;
						}
						yyn = -yyn;
						{
							parseState=yyreduce;
							continue;
						}
					}
					else if (yyn == 0)
					{
						parseState=yyerrlab;
						continue;
					}

					if (yyn == Grammar::YYFINAL)
						return 0;

					// Shift the lookahead token.
					Debug::ShiftToken(yychar, Grammar::yytname[yychar1]);

					// Discard the token being shifted unless it is eof.
					if (yychar != YYEOF)
						yychar = YYEMPTY;

					valStack.push_back(yylval);
					//locStack.push_back(yylloc);

					// count tokens shifted since error; after three, turn off error status.
					if (yyerrstatus)
						yyerrstatus--;

					yystate = yyn;
					parseState=yynewstate;
					continue;
				}
			case yydefault:
				{
					// Do the default action for the current state.
					yyn = Grammar::yydefact[yystate];
					if (yyn == 0)
					{
						parseState=yyerrlab;
						continue;
					}
				}
			case yyreduce:
				{
					// Do a reduction.  yyn is the number of a rule to reduce with.
					int yylen = Grammar::yyr2[yyn];
					if (yylen > 0)
						yyval = *(valStack.end()-yylen);

					Debug::ReduceMessage(yyn, Grammar::yyrline[yyn], Grammar::yyprhs, Grammar::yyrhs,
						Grammar::yytname, Grammar::yyr1);

					grammar.Action(yyn, yyval, &valStack.back());

					valStack.erase(valStack.end()-yylen, valStack.end());
					stateStack.erase(stateStack.end()-yylen, stateStack.end());

					valStack.push_back(yyval);

					//// Update location stack values
					//{
					//	LocationStack::value_type temp;
					//	temp.last_line = locStack.back().last_line;
					//	temp.last_column = locStack.back().last_column;

					//	if (yylen == 0)
					//	{
					//		temp.first_line = yylloc.first_line;
					//		temp.first_column = yylloc.first_column;
					//		temp.text = 0;
					//	}
					//	else
					//	{
					//		LocationStack::const_reference old = *(locStack.end()-yylen);
					//		temp.first_line = old.first_line;
					//		temp.first_column = old.first_column;
					//		temp.text = old.text;
					//	}
					//	locStack.erase(locStack.end()-yylen, locStack.end());
					//	locStack.push_back(temp);
					//}

					Debug::PrintStack(stateStack);

					//Now "shift" the result of the reduction.
					//Determine what state that goes to,
					//based on the state we popped back to
					//and the rule number reduced by.
					yyn = Grammar::yyr1[yyn];

					yystate = Grammar::yypgoto[yyn - Grammar::YYNTBASE] + stateStack.back();
					if (yystate >= 0 &&
						yystate <= Grammar::YYLAST &&
						Grammar::yycheck[yystate] == stateStack.back())
					{
						yystate = Grammar::yytable[yystate];
					}
					else
						yystate = Grammar::yydefgoto[yyn - Grammar::YYNTBASE];

					parseState=yynewstate;
					continue;
				}
			case yyerrlab:
				{
					// here on detecting error

					//If not already recovering from an error, report this error.
					if (!yyerrstatus)
						Debug::Error("parse error");

					if (yyerrstatus == 3)
					{
						// if just tried and failed to reuse lookahead token after an error, discard it.
						// return failure if at end of input
						if (yychar == YYEOF)
							return 0;

						Debug::DiscardToken(yychar, Grammar::yytname[yychar1]);

						yychar = YYEMPTY;
					}

					// Else will try to reuse lookahead token
					// after shifting the error token.

					yyerrstatus = 3; // Each real token shifted decrements this

					parseState=yyerrhandle;
					continue;
				}
			case yyerrpop:
				{
					// pop the current state because it cannot handle the error token
					if (stateStack.empty())
						return 0;

					yystate = stateStack.back();

					if (valStack.empty())
						return 0;
					//if (locStack.empty())
					//	return 0;

					stateStack.pop_back();
					valStack.pop_back();
					//locStack.pop_back();

					Debug::PrintStack(stateStack);
				}
			case yyerrhandle:
				{
					yyn = Grammar::yypact[yystate];
					if (yyn == Grammar::YYFLAG)
					{
						parseState=yyerrpop;
						continue;
					}

					yyn += YYTERROR;
					if (yyn < 0 || yyn > Grammar::YYLAST || Grammar::yycheck[yyn] != YYTERROR)
					{
						parseState=yyerrpop;
						continue;
					}

					yyn = Grammar::yytable[yyn];
					if (yyn < 0)
					{
						if (yyn == Grammar::YYFLAG)
						{
							parseState=yyerrpop;
							continue;
						}
						yyn = -yyn;
						{
							parseState=yyreduce;
							continue;
						}
					}
					else if (yyn == 0)
					{
						parseState=yyerrpop;
						continue;
					}

					if (yyn == Grammar::YYFINAL)
						return 0;

					Debug::Message("Shifting error token:");

					valStack.push_back(yylval);
					//locStack.push_back(yylloc);

					yystate = yyn;

					parseState=yynewstate;
					continue;
				}
			}
		}
	}
};

#endif //MAMMUTPARSER_H
