/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


// Find and resolve or report look-ahead conflicts

#include <stdio.h>
#include "system.h"
#include "machine.h"
#include "new.h"
#include "files.h"
#include "gram.h"
#include "state.h"
#include <malloc.h>	// for alloca()

#include <iostream>

using namespace std;

extern char **tags;
extern int tokensetsize;
extern char *consistent;
extern short *accessing_symbol;
extern shifts **shift_table;
extern unsigned *LA;
extern short *LAruleno;
extern short *lookaheads;
extern bool verboseflag;

void set_conflicts(int);
void resolve_sr_conflict(int state, int lookaheadnum);
void flush_shift(int state, int token);
void log_resolution(int state, int LAno, int token, char *resolution);
void total_conflicts();
void count_sr_conflicts(int);
void count_rr_conflicts(int);

char any_conflicts;
char *conflicts;
errs **err_table;
int expected_conflicts;

static unsigned *shiftset;
static unsigned *lookaheadset;
static int src_total;
static int rrc_total;
static int src_count;
static int rrc_count;

void initialize_conflicts()
{
	int i;
	/*  errs *sp; JF unused */
	
	conflicts = NEW2(nstates, char);
	shiftset = NEW2(tokensetsize, unsigned);
	lookaheadset = NEW2(tokensetsize, unsigned);
	
	err_table = NEW2(nstates, errs *);
	
	any_conflicts = 0;
	
	for (i = 0; i < nstates; i++)
		set_conflicts(i);
}


void set_conflicts(int state)
{
	int i;
	int k;
	shifts *shiftp;
	unsigned *fp2;
	unsigned *fp3;
	unsigned *fp4;
	unsigned *fp1;
	int symbol;
	
	if (consistent[state]) return;
	
	for (i = 0; i < tokensetsize; i++)
		lookaheadset[i] = 0;
	
	shiftp = shift_table[state];
	if (shiftp)
    {
		k = shiftp->nshifts;
		for (i = 0; i < k; i++)
		{
			symbol = accessing_symbol[shiftp->shifts[i]];
			if (ISVAR(symbol)) break;
			SETBIT(lookaheadset, symbol);
		}
    }
	
	k = lookaheads[state + 1];
	fp4 = lookaheadset + tokensetsize;
	
	/* loop over all rules which require lookahead in this state */
	/* first check for shift-reduce conflict, and try to resolve using precedence  */
	
	for (i = lookaheads[state]; i < k; i++)
		if (rprec[LAruleno[i]])
		{
			fp1 = LA + i * tokensetsize;
			fp2 = fp1;
			fp3 = lookaheadset;
			
			while (fp3 < fp4)
			{
				if (*fp2++ & *fp3++)
				{
					resolve_sr_conflict(state, i);
					break;
				}
			}
		}
		
		/* loop over all rules which require lookahead in this state */
		/* Check for conflicts not resolved above.  */
		
		for (i = lookaheads[state]; i < k; i++)
		{
			fp1 = LA + i * tokensetsize;
			fp2 = fp1;
			fp3 = lookaheadset;
			
			while (fp3 < fp4)
			{
				if (*fp2++ & *fp3++)
				{
					conflicts[state] = 1;
					any_conflicts = 1;
				}
			}
			
			fp2 = fp1;
			fp3 = lookaheadset;
			
			while (fp3 < fp4)
				*fp3++ |= *fp2++;
		}
}



/* Attempt to resolve shift-reduce conflict for one rule
by means of precedence declarations.
It has already been checked that the rule has a precedence.
A conflict is resolved by modifying the shift or reduce tables
so that there is no longer a conflict.  */

void resolve_sr_conflict(int state, int lookaheadnum)
{
	int i;
	int mask;
	unsigned *fp1;
	unsigned *fp2;
	int redprec;

	/* Extra parens avoid errors on Ultrix 4.3.  */
	errs *errp = (errs *) alloca ((sizeof(errs) + ntokens * sizeof(short)));
	short *errtokens = errp->errs;
	
	/* find the rule to reduce by to get precedence of reduction  */
	redprec = rprec[LAruleno[lookaheadnum]];
	
	mask = 1;
	fp1 = LA + lookaheadnum * tokensetsize;
	fp2 = lookaheadset;
	for (i = 0; i < ntokens; i++)
    {
		if ((mask & *fp2 & *fp1) && sprec[i])
		/* Shift-reduce conflict occurs for token number i
		and it has a precedence.
		The precedence of shifting is that of token i.  */
		{
			if (sprec[i] < redprec)
			{
				if (verboseflag) log_resolution(state, lookaheadnum, i, "reduce");
				*fp2 &= ~mask;  /* flush the shift for this token */
				flush_shift(state, i);
			}
			else if (sprec[i] > redprec)
			{
				if (verboseflag) log_resolution(state, lookaheadnum, i, "shift");
				*fp1 &= ~mask;  /* flush the reduce for this token */
			}
			else
			{
			/* Matching precedence levels.
			For left association, keep only the reduction.
			For right association, keep only the shift.
				For nonassociation, keep neither.  */
				
				switch (sassoc[i])
				{
				case RIGHT_ASSOC:
					if (verboseflag) log_resolution(state, lookaheadnum, i, "shift");
					break;
					
				case LEFT_ASSOC:
					if (verboseflag) log_resolution(state, lookaheadnum, i, "reduce");
					break;
					
				case NON_ASSOC:
					if (verboseflag) log_resolution(state, lookaheadnum, i, "an error");
					break;
				}
				
				if (sassoc[i] != RIGHT_ASSOC)
				{
					*fp2 &= ~mask;  /* flush the shift for this token */
					flush_shift(state, i);
				}
				if (sassoc[i] != LEFT_ASSOC)
				{
					*fp1 &= ~mask;  /* flush the reduce for this token */
				}
				if (sassoc[i] == NON_ASSOC)
				{
					/* Record an explicit error for this token.  */
					*errtokens++ = i;
				}
			}
		}
		
		mask <<= 1;
		if (mask == 0)
		{
			mask = 1;
			fp2++;  fp1++;
		}
    }
	errp->nerrs = errtokens - errp->errs;
	if (errp->nerrs)
    {
	/* Some tokens have been explicitly made errors.  Allocate
		a permanent errs structure for this state, to record them.  */
		i = (char *) errtokens - (char *) errp;
		err_table[state] = (errs *) calloc ((unsigned int)i,1);
		bcopy (errp, err_table[state], i);
    }
	else
		err_table[state] = 0;
}

/* turn off the shift recorded for the specified token in the specified state.
Used when we resolve a shift-reduce conflict in favor of the reduction.  */

void flush_shift(int state, int token)
{
	shifts *shiftp;
	int k, i;
	/*  unsigned symbol; JF unused */
	
	shiftp = shift_table[state];
	
	if (shiftp)
    {
		k = shiftp->nshifts;
		for (i = 0; i < k; i++)
		{
			if (shiftp->shifts[i] && token == accessing_symbol[shiftp->shifts[i]])
				(shiftp->shifts[i]) = 0;
		}
    }
}

void log_resolution(int state, int LAno, int token, char *resolution)
{
	std::clog << "Conflict in state " << state <<
		" between rule " << LAruleno[LAno] <<
		" and token " << tags[token] <<
		" resolved as " << resolution << ".\n";
}

void conflict_log()
{
	src_total = 0;
	rrc_total = 0;
	
	for (int i = 0; i < nstates; i++)
    {
		if (conflicts[i])
		{
			count_sr_conflicts(i);
			count_rr_conflicts(i);
			src_total += src_count;
			rrc_total += rrc_count;
		}
    }
	
	total_conflicts();
}

void verbose_conflict_log()
{
	src_total = 0;
	rrc_total = 0;

	for (int i = 0; i < nstates; i++)
	{
		if (conflicts[i])
		{
			count_sr_conflicts(i);
			count_rr_conflicts(i);
			src_total += src_count;
			rrc_total += rrc_count;

			if (src_count > 0)
			{
				clog << "State " << i <<
					" contains " << src_count << " shift/reduce conflicts";
				if (rrc_count > 0)
					clog << " and " << rrc_count << " reduce/reduce conflicts";
				clog << ".\n";
			}
		}
	}
	total_conflicts();
}


void total_conflicts()
{
	if (src_total == expected_conflicts && rrc_total == 0)
		return;
	
	fprintf(stderr, "%s contains", infile.c_str());
	
	if (src_total == 1)
		fprintf(stderr, " 1 shift/reduce conflict");
	else if (src_total > 1)
		fprintf(stderr, " %d shift/reduce conflicts", src_total);
	
	if (src_total > 0 && rrc_total > 0)
		fprintf(stderr, " and");
	
	if (rrc_total == 1)
		fprintf(stderr, " 1 reduce/reduce conflict");
	else if (rrc_total > 1)
		fprintf(stderr, " %d reduce/reduce conflicts", rrc_total);
	
	putc('.', stderr);
	putc('\n', stderr);
}


void count_sr_conflicts(int state)
{
	int i;
	int k;
	int mask;
	shifts *shiftp;
	unsigned *fp1;
	unsigned *fp2;
	unsigned *fp3;
	int symbol;
	
	src_count = 0;
	
	shiftp = shift_table[state];
	if (!shiftp) return;
	
	for (i = 0; i < tokensetsize; i++)
    {
		shiftset[i] = 0;
		lookaheadset[i] = 0;
    }
	
	k = shiftp->nshifts;
	for (i = 0; i < k; i++)
    {
		if (! shiftp->shifts[i]) continue;
		symbol = accessing_symbol[shiftp->shifts[i]];
		if (ISVAR(symbol)) break;
		SETBIT(shiftset, symbol);
    }
	
	k = lookaheads[state + 1];
	fp3 = lookaheadset + tokensetsize;
	
	for (i = lookaheads[state]; i < k; i++)
    {
		fp1 = LA + i * tokensetsize;
		fp2 = lookaheadset;
		
		while (fp2 < fp3)
			*fp2++ |= *fp1++;
    }
	
	fp1 = shiftset;
	fp2 = lookaheadset;
	
	while (fp2 < fp3)
		*fp2++ &= *fp1++;
	
	mask = 1;
	fp2 = lookaheadset;
	for (i = 0; i < ntokens; i++)
    {
		if (mask & *fp2)
			src_count++;
		
		mask <<= 1;
		if (mask == 0)
		{
			mask = 1;
			fp2++;
		}
    }
}


void count_rr_conflicts(int state)
{
	int i;
	int j;
	int count;
	unsigned mask;
	unsigned *baseword;
	unsigned *wordp;
	int m;
	int n;
	
	rrc_count = 0;
	
	m = lookaheads[state];
	n = lookaheads[state + 1];
	
	if (n - m < 2) return;
	
	mask = 1;
	baseword = LA + m * tokensetsize;
	for (i = 0; i < ntokens; i++)
    {
		wordp = baseword;
		
		count = 0;
		for (j = m; j < n; j++)
		{
			if (mask & *wordp)
				count++;
			
			wordp += tokensetsize;
		}
		
		if (count >= 2) rrc_count++;
		
		mask <<= 1;
		if (mask == 0)
		{
			mask = 1;
			baseword++;
		}
    }
}

void log_helper(char* tag, int default_rule, char* rule)
{
	clog << "    " <<
		tag << '\t' <<
		"[reduce using rule " << default_rule << " (" << rule << ")]\n";
}

void print_reductions(int state)
{
	int i;
	int j;
	int k;
	unsigned *fp1;
	unsigned *fp2;
	unsigned *fp3;
	unsigned *fp4;
	int rule;
	int symbol;
	unsigned mask;
	int m;
	int n;
	int default_LA;
	int default_rule;
	int cmax;
	int count;
	shifts *shiftp;
	errs *errp;
	int nodefault = 0;
	
	for (i = 0; i < tokensetsize; i++)
		shiftset[i] = 0;
	
	shiftp = shift_table[state];
	if (shiftp)
    {
		k = shiftp->nshifts;
		for (i = 0; i < k; i++)
		{
			if (! shiftp->shifts[i]) continue;
			symbol = accessing_symbol[shiftp->shifts[i]];
			if (ISVAR(symbol)) break;
			/* if this state has a shift for the error token,
			don't use a default rule.  */
			if (symbol == error_token_number) nodefault = 1;
			SETBIT(shiftset, symbol);
		}
    }
	
	errp = err_table[state];
	if (errp)
    {
		k = errp->nerrs;
		for (i = 0; i < k; i++)
		{
			if (! errp->errs[i]) continue;
			symbol = errp->errs[i];
			SETBIT(shiftset, symbol);
		}
    }
	
	m = lookaheads[state];
	n = lookaheads[state + 1];
	
	if (n - m == 1 && ! nodefault)
    {
		default_rule = LAruleno[m];
		
		fp1 = LA + m * tokensetsize;
		fp2 = shiftset;
		fp3 = lookaheadset;
		fp4 = lookaheadset + tokensetsize;
		
		while (fp3 < fp4)
			*fp3++ = *fp1++ & *fp2++;
		
		mask = 1;
		fp3 = lookaheadset;
		
		for (i = 0; i < ntokens; i++)
		{
			if (mask & *fp3)
				log_helper(tags[i], default_rule, tags[rlhs[default_rule]]);
			
			mask <<= 1;
			if (mask == 0)
			{
				mask = 1;
				fp3++;
			}
		}
		
		clog << "    $default\treduce using rule " << default_rule << " (" << tags[rlhs[default_rule]] << ")\n\n";
    }
	else if (n - m >= 1)
    {
		cmax = 0;
		default_LA = -1;
		fp4 = lookaheadset + tokensetsize;
		
		if (! nodefault)
			for (i = m; i < n; i++)
			{
				fp1 = LA + i * tokensetsize;
				fp2 = shiftset;
				fp3 = lookaheadset;
				
				while (fp3 < fp4)
					*fp3++ = *fp1++ & ( ~ (*fp2++));
				
				count = 0;
				mask = 1;
				fp3 = lookaheadset;
				for (j = 0; j < ntokens; j++)
				{
					if (mask & *fp3)
						count++;
					
					mask <<= 1;
					if (mask == 0)
					{
						mask = 1;
						fp3++;
					}
				}
				
				if (count > cmax)
				{
					cmax = count;
					default_LA = i;
					default_rule = LAruleno[i];
				}
				
				fp2 = shiftset;
				fp3 = lookaheadset;
				
				while (fp3 < fp4)
					*fp2++ |= *fp3++;
			}
			
			for (i = 0; i < tokensetsize; i++)
				shiftset[i] = 0;
			
			if (shiftp)
			{
				k = shiftp->nshifts;
				for (i = 0; i < k; i++)
				{
					if (! shiftp->shifts[i]) continue;
					symbol = accessing_symbol[shiftp->shifts[i]];
					if (ISVAR(symbol)) break;
					SETBIT(shiftset, symbol);
				}
			}
			
			mask = 1;
			fp1 = LA + m * tokensetsize;
			fp2 = shiftset;
			for (i = 0; i < ntokens; i++)
			{
				int defaulted = 0;
				
				if (mask & *fp2)
					count = 1;
				else
					count = 0;
				
				fp3 = fp1;
				for (j = m; j < n; j++)
				{
					if (mask & *fp3)
					{
						if (count == 0)
						{
							if (j != default_LA)
							{
								rule = LAruleno[j];
								log_helper(tags[i], rule, tags[rlhs[rule]]);
							}
							else defaulted = 1;
							
							count++;
						}
						else
						{
							if (defaulted)
							{
								rule = LAruleno[default_LA];
								log_helper(tags[i], rule, tags[rlhs[rule]]);
								defaulted = 0;
							}
							rule = LAruleno[j];
							log_helper(tags[i], rule, tags[rlhs[rule]]);
						}
					}
					
					fp3 += tokensetsize;
				}
				
				mask <<= 1;
				if (mask == 0)
				{
					mask = 1;
					/* This used to be fp1, but I think fp2 is right
					because fp2 is where the words are fetched to test with mask
					in this loop.  */
					fp2++;
				}
			}
			
			if (default_LA >= 0)
				clog << "    $default\treduce using rule " << default_rule << " (" << tags[rlhs[default_rule]] << ")\n";
			
			clog.put('\n');
    }
}


void finalize_conflicts()
{
	FREE(conflicts);
	FREE(shiftset);
	FREE(lookaheadset);
}
