/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


#include <stdio.h>
#include "system.h"
#include "machine.h"
#include <string>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include "Config.h"
#include "symtab.h"
#include "files.h"
using namespace std;

extern	int lineno;
extern	string infile;
extern char *version_string;
extern bucket *firstsymbol;

bool verboseflag;
bool nolinesflag;
bool noconfig;

/* Nonzero means failure has been detected; don't write a parser file.  */
int failure;

extern void openfiles(), reader(), reduce_grammar();
extern void set_derives(), set_nullable(), generate_states();
extern void lalr(), initialize_conflicts(), verbose(), terse();
extern void CreateCode(map<string, string> &codeMap);
extern void output(map<string, string> &codeMap), done(int);
extern void output_about_token();
extern string computeDefaultBase();
extern bool parser_name_defined;

#include <boost/program_options/options_description.hpp>
#include <boost/program_options/positional_options.hpp>
#include <boost/program_options/variables_map.hpp>
#include <boost/program_options/parsers.hpp>
namespace po = boost::program_options;

#include <boost/filesystem/path.hpp>
namespace fs = boost::filesystem;

void ExecuteCommand(map<string, string> &codeMap);
void ExecuteConfig(po::parsed_options const&, map<string, string> &codeMap, fs::path const& path);
void AddKeywords(string const& keywords, map<string, string> & codeMap);

po::options_description desc("Allowed options");
po::positional_options_description p;

string keywords;

int main(int argc, char **argv)
{
#ifdef WIN32
	char exeName[MAX_PATH];
	::GetModuleFileName(NULL, exeName, sizeof(exeName));
	fs::path progPath = fs::path(exeName, fs::native).branch_path();
#else
	fs::path progPath = fs::path(argv[0], fs::native).branch_path();

	if(progPath.empty())
		progPath = ".";
#endif

	// TODO: add another option that allows for adding config-file keywords
	// add another option that allows for adding template keywords
	desc.add_options()
		("help", "Shows information about options.")
		("nolines,l", "Do not output #lines to the output.")
		("noconfig,n", "Do not use the default config file.")
		("verbose,v", "Verbose message output.")
		("version,V", "Prints out the version number.")
		("config,c", po::value<string>(&configuration_file), "Sets config file.")
		("prefix,p", po::value<string>(&file_prefix), "Output files prefix. In conjunction with --config.")
		("keywords,k", po::value<string>(&keywords), "Define keywords reachable from the templates [key=value[,key=value]].")
		("output,o", po::value<string>(&output_file), "Generated output file name.")
		("template,t", po::value<string>(&output_template), "The template for the output file.")
		("input", po::value<string>(&infile), "Input grammar file.")
		;
	p.add("input", -1);
	po::variables_map vm;

	try {
		failure = 0;
		lineno = 0;
		// Get initial arguments (e.g. configuration-filename)
		po::parsed_options base_opts = po::command_line_parser(argc, argv).options(desc).positional(p).run();
		po::store(base_opts, vm);
		po::notify(vm);

		if (vm.count("help"))
		{
			string const progName = fs::path(argv[0], fs::native).leaf();
			cout << "Usage: " << progName << " [options] [input file]\n\n";
			cout << desc << endl;
			cout << "Example: " << progName << " -p MyPrefix Grammar.y\n\n";
			cout << "This will use the default config file and produce MyPrefix.h and MyPrefix.cpp from the Grammmar.y file.\n" << endl;
			return 0;
		}
		if (vm.count("version"))
		{
			cout << version_string << endl;
			return 0;
		}
		verboseflag = vm.count("verbose") ? true : false;
		nolinesflag = vm.count("nolines") ? true : false;
		noconfig = vm.count("noconfig") ? true : false;
		
		openfiles();
		/* read the input.  Copy some parts of it to faction, ftable.
		In file reader.c.
		The other parts are recorded in the grammar; see gram.h.  */
		reader();
		
		/* find useless nonterminals and productions and reduce the grammar.  In
		file reduce.c */
		reduce_grammar();
		
		/* record other info about the grammar.  In files derives and nullable.  */
		set_derives();
		set_nullable();
		
		/* convert to nondeterministic finite state machine.  In file LR0.
		See state.h for more info.  */
		generate_states();
		
		/* make it deterministic.  In file lalr.  */
		lalr();
		
		/* Find and record any conflicts: places where one token of lookahead is not
		enough to disambiguate the parsing.  In file conflicts.
		Currently this does not do anything to resolve them;
		the trivial form of conflict resolution that exists is done in output.  */
		initialize_conflicts();
		
		/* print information about results, if requested.  In file print. */
		if (verboseflag)
			verbose();
		else
			terse();

		map<string, string> codeMap;
		CreateCode(codeMap);
		if (!keywords.empty())
			AddKeywords(keywords, codeMap);
		keywords = "";

		if (noconfig)
			ExecuteCommand(codeMap);
		else
			ExecuteConfig(base_opts, codeMap, progPath);
	}
	catch(exception &e)
	{
		if (!infile.empty())
			cout << infile << " : ";
		cout << "error: " << e.what() << std::endl;
	}
	done(failure);
}

void AddKeywords(string const& keywords, map<string, string> & codeMap)
{
	string key, value;
	stringstream ss(keywords);
	while(ss.good())
	{
		getline(ss, key, '=');
		getline(ss, value, ',');
		codeMap[key] = value;
	}
}

void ExecuteConfig(po::parsed_options const& base_opts, map<string, string> &codeMap,
	fs::path const& cfgPath)
{
	// TODO: The loading of the configuration could be nicer, but works for now.
	Configuration conf;
	if (file_prefix.empty())
	{
		file_prefix = infile;
	}
	conf.LoadKeyword("file_prefix", file_prefix);

	if (!output_template.empty())
		conf.LoadKeyword("output_template", output_template);

	// Use the default config file if none is specified
	if (configuration_file.empty())
		configuration_file = (cfgPath / fs::path("configfiles/mammut.conf")).native_file_string();

	conf.LoadKeyword("configuration_file", configuration_file);
	conf.LoadKeyword("program_dir", cfgPath.string());

	ifstream file(configuration_file.c_str());
	if (!file)
		throw std::runtime_error("Unable to open configuration file: " + configuration_file);

	// Read one line at a time from the configuration file and execute mammut for each one of these
	while(file.good())
	{
		string line;
		ws(file);
		getline(file, line);

		// Skip empty lines
		if (line.empty())
			continue;

		vector<string> args = conf.ParseCommandLine(line);

		if (args.empty())
			continue;

		po::variables_map vm;
		po::store(po::command_line_parser(args).options(desc).positional(p).run(), vm);
		// The manually supplied arguments take precedence, load them last
		po::store(base_opts, vm);
		po::notify(vm);

		if (!keywords.empty())
		{
			AddKeywords(keywords, codeMap);
			keywords.clear();
		}

		ExecuteCommand(codeMap);
	}
}
void ExecuteCommand(map<string, string> &codeMap)
{	
	/* output the tables and the parser to ftable.  In file output. */
	if (output_file.empty())
		throw std::runtime_error("No output file was defined");
	if (output_template.empty())
		throw std::runtime_error("No template file was defined");

	output(codeMap);
}

/* functions to report errors which prevent a parser from being generated */
void fatal(char *s)
{
	if (infile.empty())
		cerr << "fatal error: " << s << endl;
	else
		cerr << infile << '(' << lineno << ") : error: " << s << endl;

	done(1);
}

void fatals(char const*fmt, char const*x1, char const*x2, char const*x3)
{
	char buffer[200];
	sprintf(buffer, fmt, x1,x2,x3);
	fatal(buffer);
}

void toomany(char *s)
{
	char buffer[200];
	sprintf(buffer, "limit of %d exceeded, too many %s", MAXSHORT, s);
	fatal(buffer);
}

void berror(char *s)
{
	cerr << "internal error, " << s << endl;
	done(1);
}
