/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


/* Output the generated parsing program */

/* functions to output parsing data to various files.  Entries are:

output_headers ()
Output constant strings to the beginning of certain files.
output_trailers()
Output constant strings to the ends of certain files.
output ()
Output the parsing tables and the parser code to ftable.
The parser tables consist of these tables.
tarred ones needed only for the semantic parser.
yytranslate = vector mapping yylex's token numbers into bison's token numbers.
yytname = vector of string-names indexed by bison token number
yyrline = vector of line-numbers of all rules.  For yydebug printouts.
yyrhs = vector of items of all rules.
This is exactly what ritems contains.

yyprhs[r] = index in yyrhs of first item for rule r.

yyr1[r] = symbol number of symbol that rule r derives.

yyr2[r] = number of symbols composing right hand side of rule r.

yydefact[s] = default rule to reduce with in state s,
when yytable doesn't specify something else to do.
Zero means the default is an error.

yydefgoto[i] = default state to go to after a reduction of a rule that
generates variable ntokens + i, except when yytable
specifies something else to do.

yypact[s] = index in yytable of the portion describing state s.
The lookahead token's type is used to index that portion
to find out what to do.

If the value in yytable is positive,
we shift the token and go to that state.
If the value is negative, it is minus a rule number to reduce by.
If the value is zero, the default action from yydefact[s] is used.

yypgoto[i] = the index in yytable of the portion describing 
what to do after reducing a rule that derives variable i + ntokens.
This portion is indexed by the parser state number
as of before the text for this nonterminal was read.
The value from yytable is the state to go to.

yytable = a vector filled with portions for different uses,
found via yypact and yypgoto.

yycheck = a vector indexed in parallel with yytable.
It indicates, in a roundabout way, the bounds of the
portion you are trying to examine.

Suppose that the portion of yytable starts at index p
and the index to be examined within the portion is i.
Then if yycheck[p+i] != i, i is outside the bounds
of what is actually allocated, and the default
(from yydefact or yydefgoto) should be used.
Otherwise, yytable[p+i] should be used.

YYFINAL = the state number of the termination state.
YYFLAG = most negative short int.  Used to flag ??
YYNTBASE = ntokens.
*/

#include <iomanip>
#include <sstream>
#include <map>
#include <stdexcept>

using namespace std;

#include <boost/filesystem/path.hpp>
namespace fs = boost::filesystem;

#include <stdio.h>
#include "system.h"
#include "machine.h"
#include "new.h"
#include "files.h"
#include "gram.h"
#include "state.h"
#include "symtab.h"

stringstream uniondeclaration;

// TODO: remove the definition of headerdeclaration and implementationdeclarations.
// These should be dynamically created as needed instead.
stringstream headerdeclaration;
stringstream implementationdeclaration;

extern char **tags;
extern int tokensetsize;
extern int final_state;
extern core **state_table;
extern shifts **shift_table;
extern errs **err_table;
extern reductions **reduction_table;
extern short *accessing_symbol;
extern unsigned *LA;
extern short *LAruleno;
extern short *lookaheads;
extern char *consistent;
extern short *goto_map;
extern short *from_state;
extern short *to_state;

extern void free_symtab();

void sort_actions();
void pack_table();
void free_shifts();
void free_reductions();
void free_itemsets();

int action_row(int state);
void save_row(int state);
int default_goto(int ntokens);
void save_column(int symbol, int default_state);
int matching_state(int vector);
int pack_vector(int vector);
void berror(char *s);
void fatals(char const*fmt, char const*x1, char const*x2, char const*x3);

static int nvectors;
static int nentries;
static short **froms;
static short **tos;
static short *tally;
static short *width;
static short *actrow;
static short *state_count;
static short *order;
static short *base;
static short *pos;
static short *table;
static short *check;
static int lowzero;
static int high;

void EmitTokenEnum(ostream &os);
void EmitTokenDefinesFmt(ostream &os, char *fmt, int notrans);
extern bucket *errtoken;
void EmitTokenDefines();
void EmitTokenConstDef();
void EmitTokenConstDecl();

extern string output_template;
extern char *version_string;

#ifdef _MSC_VER
 #pragma warning(disable : 4244)
#endif

class LineBreaker
{
public:
	LineBreaker(ostream &os, int width = 10)
		: m_os(os), m_col(0), m_width(width), m_first(true)
	{
	}

	template<typename T>
	void operator <<(T const& val)
	{
		if(m_first)
			m_first = false;
		else
		{
			m_os << ", ";
			if(m_col >= m_width)
			{
				m_col = 0;
				m_os << "\n";
			}
		}

		m_os << setw(6) << val;
		m_col++;
	}

private:
	LineBreaker &operator=(LineBreaker const &) { return *this; }

protected:
	ostream &m_os;
	int m_col, m_width;
	bool m_first;
};

static void StoreCode(string const& name, void (*emit)(ostream &), map<string, string> &codeMap)
{
	stringstream ss;
	emit(ss);
	codeMap[name] = ss.str();
}

//static void CopyFileToStream(FILE* file, ostream & os)
//{
//	rewind(file);
//	int c;
//	while ((c = getc(file)) != EOF)
//		os.put(static_cast<char>(c));
//}

void EmitDefines(ostream & os)
{
	os << "YYFINAL=" << final_state;
	os << ", YYFLAG=" << MINSHORT;
	os << ", YYNTBASE=" << ntokens;
	os << ", YYLAST=" << high;
}

void EmitMaxtokennumber(ostream & os)
{
	os << max_user_token_number;
}

void EmitNsyms(ostream & os)
{
	os << nsyms;
}

void EmitTokenTranslations(ostream & os)
{
	LineBreaker lb(os);
	lb << 0;
	
	for (int i=1; i<=max_user_token_number; i++)
		lb << token_translations[i];
}

void EmitPrhs(ostream & os)
{
	LineBreaker lb(os);
	lb << 0;
		
	for (int i = 1; i <= nrules; i++)
		lb << rrhs[i];
}

void EmitRline(ostream &os)
{
	LineBreaker lb(os);
	lb << 0;

	for (int i = 1; i <= nrules; i++)
		lb << rline[i];
}

void EmitR1(ostream &os)
{
	LineBreaker lb(os);
	lb << 0;
	for (int i = 1; i <= nrules; i++)
		lb << rlhs[i];

	FREE(rlhs + 1);
}

void EmitR2(ostream &os)
{
	LineBreaker lb(os);
	lb << 0;
	for (int i = 1; i < nrules; i++)
		lb << rrhs[i + 1] - rrhs[i] - 1;

	lb << nitems - rrhs[nrules] - 1;
		
	FREE(rrhs + 1);
}

void EmitRhs(ostream & os)
{
	LineBreaker lb(os);
	for (short const* sp = ritem; *sp; sp++)
		lb << *sp;
}


void EmitTname(ostream &os)
{
	LineBreaker lb(os, 5);
	for (int i = 0; i <= nsyms; i++)
	{
		stringstream ss;
		
		ss << '\"';
		
		for (char *p = tags[i]; p && *p; p++)
		{
			if (*p < 040 || *p >= 0177)
				ss << "\\" << setw(3) << setfill('0') << *p << "o";
			else
			{
				switch(*p)
				{
				case '"':
				case '\\': ss << "\\" << *p;
					break;
				case '\n': ss << "\\n";
					break;
				case '\t': ss << "\\t";
					break;
				case '\b': ss << "\\b";
					break;
				default: ss << *p;
					break;
				};
			}
		}
		
		ss << '\"';
		lb << ss.str();
	}
}

/*	Figure out the actions for the specified state, indexed by lookahead token type.
	The yydefact table is output now.  The detailed info
	is saved for putting into yytable later.  */

void EmitTokenActions(ostream &os)
{
	actrow = NEW2(ntokens, short);
	
	LineBreaker lb(os);
	for (int i = 0; i < nstates; i++)
    {
		lb << action_row(i);
		save_row(i);
    }
	
	FREE(actrow);
}

/* Figure out what to do after reducing with each rule,
depending on the saved state from before the beginning
of parsing the data that matched this rule.

  The yydefgoto table is output now.  The detailed info
is saved for putting into yytable later.  */

void EmitGotoActions(ostream &os)
{
	state_count = NEW2(nstates, short);
	
	LineBreaker lb(os);
	for (int i = ntokens; i < nsyms; i++)
    {
		int const k = default_goto(i);
		lb << k;
		save_column(i, k);
    }
	
	FREE(state_count);
}

/* the following functions output yytable, yycheck
and the vectors whose elements index the portion starts */

void EmitPact(ostream &os)
{
	LineBreaker lb(os);
	for (int i = 0; i < nstates; i++)
		lb << base[i];
}

void EmitPgoto(ostream &os)
{
	LineBreaker lb(os);
	for (int i = nstates; i < nvectors; i++)
		lb << base[i];
	FREE(base);
}

void EmitTable(ostream &os)
{
	LineBreaker lb(os);
	for (int i = 0; i <= high; i++)
		lb << table[i];

	FREE(table);
}

void EmitCheck(ostream &os)
{
	LineBreaker lb(os);
	for (int i = 0; i <= high; i++)
		lb << check[i];

	FREE(check);
}

void EmitActions(map<string, string> &codeMap)
{
	nvectors = nstates + nvars;
	
	froms = NEW2(nvectors, short *);
	tos = NEW2(nvectors, short *);
	tally = NEW2(nvectors, short);
	width = NEW2(nvectors, short);

	StoreCode("defact", EmitTokenActions, codeMap);

	free_shifts();
	free_reductions();
	FREE(lookaheads);
	FREE(LA);
	FREE(LAruleno);
	FREE(accessing_symbol);
	
	StoreCode("defgoto", EmitGotoActions, codeMap);

	FREE(goto_map + ntokens);
	FREE(from_state);
	FREE(to_state);
	
	sort_actions();
	pack_table();

	StoreCode("pact", EmitPact, codeMap);
	StoreCode("pgoto", EmitPgoto, codeMap);
	StoreCode("table", EmitTable, codeMap);
	StoreCode("check", EmitCheck, codeMap);
}

void CreateCode(map<string, string> &codeMap)
{
	codeMap["parsername"] = parser_name;
	codeMap["headerfilename"] = output_file;

	codeMap["union"] = uniondeclaration.str();
	codeMap["header"] = headerdeclaration.str();
	codeMap["implementation"] = implementationdeclaration.str();

	StoreCode("enums", EmitTokenEnum, codeMap);

	
	// Free the symbol table data structure since symbols are now all referred to by symbol number
	free_symtab();
	free_itemsets();
	

	StoreCode("maxtokennumber", EmitMaxtokennumber, codeMap);
	StoreCode("nsyms", EmitNsyms, codeMap);
	StoreCode("translate", EmitTokenTranslations, codeMap);
	StoreCode("prhs", EmitPrhs, codeMap);
	StoreCode("rhs", EmitRhs, codeMap);
	
	FREE(ritem);
	
	
	StoreCode("rline", EmitRline, codeMap);
	StoreCode("tname", EmitTname, codeMap);
	StoreCode("r1", EmitR1, codeMap);
	StoreCode("r2", EmitR2, codeMap);
	EmitActions(codeMap);

	codeMap["actions"] = faction.str();

	StoreCode("defines", EmitDefines, codeMap);
}

void ExpandKeywords(istream &is, map<string, string> const& codeMap, ostream &os)
{
	// TODO: output line numbers
	string line;
	while(!!getline(is, line))
	{
		string::size_type const dollar = line.find('$');
		if(dollar != string::npos)
		{
			os << string(line.begin(), line.begin() + dollar);
			string::const_iterator iter = line.begin() + dollar + 1;
			
			// TODO: handle multiple "$ sections" in one line, also make them terminatable
			// a la $keyword$ to be able to concatenate words efficiently
			// (use $$ to create a $ in the output file)
			if (isalnum(*iter))
			{
				string::const_iterator identifierEnd = iter + 1;
				while(isalnum(*identifierEnd))
					identifierEnd++;
				string word(iter, identifierEnd);

				map<string, string>::const_iterator entry = codeMap.find(word);
				if (entry != codeMap.end())
					os << (*entry).second;
				else
					os << "\n#error \"Unknown keyword \"$" << word << "\".\n";

				iter = identifierEnd;
			}
			else if (*iter == '$')
				iter++;

			// Take special care here not to upset GCC:s template matching facilities...
			string::const_iterator const outStart = iter, outEnd = line.end();
			os << string(outStart, outEnd) << '\n';
		}
		else
			os << line << '\n';
	}
}

void output(map<string, string> &codeMap)
{
#ifdef DUMP_CODEMAP
	map<string,string>::const_iterator it = codeMap.begin();

	for(; it != codeMap.end(); it++)
	{
		OutputDebugString("// $");
		OutputDebugString((*it).first.c_str());
		OutputDebugString(":\n");

		OutputDebugString((*it).second.c_str());
		OutputDebugString("\n");
	}
#endif

	{
		ifstream is(output_template.c_str());
		if (!is.is_open())
			throw runtime_error("Unable to open template file '" + output_template + "'.");

		ofstream os(output_file.c_str());
		if (!os.is_open())
			throw runtime_error("Unable to open output file '" + output_file + "'.");

		ExpandKeywords(is, codeMap, os);
	}
}

/* Decide what to do for each type of token if seen as the lookahead token in specified state.
The value returned is used as the default action (yydefact) for the state.
In addition, actrow is filled with what to do for each kind of token,
index by symbol number, with zero meaning do the default action.
The value MINSHORT, a very negative number, means this situation
is an error.  The parser recognizes this value specially.

This is where conflicts are resolved.  The loop over lookahead rules
considered lower-numbered rules last, and the last rule considered that likes
a token gets to handle it.  */

int action_row(int state)
{
	int i, j, k, m, n;
	int count;
	int default_rule;
	int nreds;
	int max;
	int rule;
	int shift_state;
	int symbol;
	unsigned mask;
	unsigned *wordp;
	reductions *redp;
	shifts *shiftp;
	errs *errp;
	int nodefault = 0;  /* set nonzero to inhibit having any default reduction */
	
	for (i = 0; i < ntokens; i++)
		actrow[i] = 0;
	
	default_rule = 0;
	nreds = 0;
	redp = reduction_table[state];
	
	if (redp)
    {
		nreds = redp->nreds;
		
		if (nreds >= 1)
		{
			/* loop over all the rules available here which require lookahead */
			m = lookaheads[state];
			n = lookaheads[state + 1];
			
			for (i = n - 1; i >= m; i--)
			{
				rule = - LAruleno[i];
				wordp = LA + i * tokensetsize;
				mask = 1;
				
				/* and find each token which the rule finds acceptable to come next */
				for (j = 0; j < ntokens; j++)
				{
					/* and record this rule as the rule to use if that token follows.  */
					if (mask & *wordp)
						actrow[j] = rule;
					
					mask <<= 1;
					if (mask == 0)
					{
						mask = 1;
						wordp++;
					}
				}
			}
		}
    }
	
	shiftp = shift_table[state];
	
	/* now see which tokens are allowed for shifts in this state.
	For them, record the shift as the thing to do.  So shift is preferred to reduce.  */
	
	if (shiftp)
    {
		k = shiftp->nshifts;
		
		for (i = 0; i < k; i++)
		{
			shift_state = shiftp->shifts[i];
			if (! shift_state) continue;
			
			symbol = accessing_symbol[shift_state];
			
			if (ISVAR(symbol))
				break;
			
			actrow[symbol] = shift_state;
			
			/* do not use any default reduction if there is a shift for error */
			
			if (symbol == error_token_number) nodefault = 1;
		}
    }
	
	errp = err_table[state];
	
	/* See which tokens are an explicit error in this state
	(due to %nonassoc).  For them, record MINSHORT as the action.  */
	
	if (errp)
    {
		k = errp->nerrs;
		
		for (i = 0; i < k; i++)
		{
			symbol = errp->errs[i];
			actrow[symbol] = (short)MINSHORT;
		}
    }
	
	/* now find the most common reduction and make it the default action for this state.  */
	
	if (nreds >= 1 && ! nodefault)
    {
		if (consistent[state])
			default_rule = redp->rules[0];
		else
		{
			max = 0;
			for (i = m; i < n; i++)
			{
				count = 0;
				rule = - LAruleno[i];
				
				for (j = 0; j < ntokens; j++)
				{
					if (actrow[j] == rule)
						count++;
				}
				
				if (count > max)
				{
					max = count;
					default_rule = rule;
				}
			}
			
			/* actions which match the default are replaced with zero,
			which means "use the default" */
			
			if (max > 0)
			{
				for (j = 0; j < ntokens; j++)
				{
					if (actrow[j] == default_rule)
						actrow[j] = 0;
				}
				
				default_rule = - default_rule;
			}
		}
    }
	
	/* If have no default rule, the default is an error.
	So replace any action which says "error" with "use default".  */
	
	if (default_rule == 0)
		for (j = 0; j < ntokens; j++)
		{
			if (actrow[j] == MINSHORT)
				actrow[j] = 0;
		}
		
	return default_rule;
}


void save_row(int state)
{
	int i;
	int count;
	short *sp;
	short *sp1;
	short *sp2;
	
	count = 0;
	for (i = 0; i < ntokens; i++)
    {
		if (actrow[i] != 0)
			count++;
    }
	
	if (count == 0)
		return;
	
	froms[state] = sp1 = sp = NEW2(count, short);
	tos[state] = sp2 = NEW2(count, short);
	
	for (i = 0; i < ntokens; i++)
    {
		if (actrow[i] != 0)
		{
			*sp1++ = i;
			*sp2++ = actrow[i];
		}
    }
	
	tally[state] = count;
	width[state] = sp1[-1] - sp[0] + 1;
}

int default_goto(int symbol)
{
	int i;
	int m;
	int n;
	int default_state;
	int max;
	
	m = goto_map[symbol];
	n = goto_map[symbol + 1];
	
	if (m == n)
		return (-1);
	
	for (i = 0; i < nstates; i++)
		state_count[i] = 0;
	
	for (i = m; i < n; i++)
		state_count[to_state[i]]++;
	
	max = 0;
	default_state = -1;
	
	for (i = 0; i < nstates; i++)
    {
		if (state_count[i] > max)
		{
			max = state_count[i];
			default_state = i;
		}
    }
	
	return (default_state);
}


void save_column(int symbol, int default_state)
{
	int m = goto_map[symbol];
	int n = goto_map[symbol + 1];
	int count = 0;

	int i;
	for (i = m; i < n; i++)
    {
		if (to_state[i] != default_state)
			count++;
    }
	
	if (count == 0)
		return;
	
	int symno = symbol - ntokens + nstates;
	
	short *sp;
	short *sp1;
	short *sp2;
	froms[symno] = sp1 = sp = NEW2(count, short);
	tos[symno] = sp2 = NEW2(count, short);
	
	for (i = m; i < n; i++)
    {
		if (to_state[i] != default_state)
		{
			*sp1++ = from_state[i];
			*sp2++ = to_state[i];
		}
    }
	
	tally[symno] = count;
	width[symno] = sp1[-1] - sp[0] + 1;
}


/* the next few functions decide how to pack 
the actions and gotos information into yytable. */

void sort_actions()
{
	order = NEW2(nvectors, short);
	nentries = 0;
	
	for (int i = 0; i < nvectors; i++)
    {
		if (tally[i] > 0)
		{
			int t = tally[i];
			int w = width[i];
			int j = nentries - 1;
			
			while (j >= 0 && (width[order[j]] < w))
				j--;
			
			while (j >= 0 && (width[order[j]] == w) && (tally[order[j]] < t))
				j--;
			
			for (int k = nentries - 1; k > j; k--)
				order[k + 1] = order[k];
			
			order[j + 1] = i;
			nentries++;
		}
    }
}

void pack_table()
{
	int i;
	int place;
	int state;
	
	base = NEW2(nvectors, short);
	pos = NEW2(nentries, short);
	table = NEW2(MAXTABLE, short);
	check = NEW2(MAXTABLE, short);
	
	lowzero = 0;
	high = 0;
	
	for (i = 0; i < nvectors; i++)
		base[i] = (short)MINSHORT;
	
	for (i = 0; i < MAXTABLE; i++)
		check[i] = -1;
	
	for (i = 0; i < nentries; i++)
    {
		state = matching_state(i);
		
		if (state < 0)
			place = pack_vector(i);
		else
			place = base[state];
		
		pos[i] = place;
		base[order[i]] = place;
    }
	
	for (i = 0; i < nvectors; i++)
    {
		if (froms[i])
			FREE(froms[i]);
		if (tos[i])
			FREE(tos[i]);
    }
	
	FREE(froms);
	FREE(tos);
	FREE(pos);
}

int matching_state(int vector)
{
	int i = order[vector];
	if (i >= nstates)
		return (-1);
	
	int t = tally[i];
	int w = width[i];
	
	for (int prev = vector - 1; prev >= 0; prev--)
    {
		int j = order[prev];
		if (width[j] != w || tally[j] != t)
			return (-1);
		
		int match = 1;
		for (int k = 0; match && k < t; k++)
		{
			if (tos[j][k] != tos[i][k] || froms[j][k] != froms[i][k])
				match = 0;
		}
		
		if (match)
			return (j);
    }
	
	return (-1);
}

int pack_vector(int vector)
{
	int i;
	int j;
	int k;
	int t;
	int loc;
	int ok;
	short *from;
	short *to;
	
	i = order[vector];
	t = tally[i];
	
	if (t == 0)
		berror("pack_vector");
	
	from = froms[i];
	to = tos[i];
	
	for (j = lowzero - from[0]; j < MAXTABLE; j++)
    {
		ok = 1;
		
		for (k = 0; ok && k < t; k++)
		{
			loc = j + from[k];
			if (loc > MAXTABLE)
				fatals("maximum table size (%d) exceeded", (char *)MAXTABLE, 0, 0);
			
			if (table[loc] != 0)
				ok = 0;
		}
		
		for (k = 0; ok && k < vector; k++)
		{
			if (pos[k] == j)
				ok = 0;
		}
		
		if (ok)
		{
			for (k = 0; k < t; k++)
			{
				loc = j + from[k];
				table[loc] = to[k];
				check[loc] = from[k];
			}
			
			while (table[lowzero] != 0)
				lowzero++;
			
			if (loc > high)
				high = loc;
			
			return (j);
		}
    }
	
	berror("pack_vector");
	return 0;	/* JF keep lint happy */
}

void free_itemsets()
{
	core *cp,*cptmp;
	
	FREE(state_table);
	
	for (cp = first_state; cp; cp = cptmp) {
		cptmp=cp->next;
		FREE(cp);
	}
}

void free_shifts()
{
	shifts *sp,*sptmp;/* JF derefrenced freed ptr */
	
	FREE(shift_table);
	
	for (sp = first_shift; sp; sp = sptmp) {
		sptmp=sp->next;
		FREE(sp);
	}
}


void free_reductions()
{
	reductions *rp,*rptmp;/* JF fixed freed ptr */
	
	FREE(reduction_table);
	
	for (rp = first_reduction; rp; rp = rptmp) {
		rptmp=rp->next;
		FREE(rp);
	}
}

void EmitTokenEnum(ostream &os)
{
	EmitTokenDefinesFmt(os, "%s=%d",0);
}

void EmitTokenDefinesFmt(ostream &os, char *fmt, int notrans)
{
	char tokenBuf[200];
	LineBreaker lb(os,10);
	for (bucket *bp = firstsymbol; bp != 0; bp = bp->next)
    {
		if (bp->value >= ntokens)
			continue;
		
		// For named tokens, but not literal ones, define the name.
		// The value is the user token number.
		if ('\'' != *tags[bp->value] && bp != errtoken)
		{
			char *cp = tags[bp->value];
			char c;
			
			/* Don't #define nonliteral tokens whose names contain periods.  */
			
			while ((c = *cp++) && c != '.')
				;

			if (!c)
			{
				sprintf(tokenBuf, fmt, tags[bp->value],
					(translations && !notrans ? bp->user_token_number : bp->value));
				lb << tokenBuf;
			}
		}
    }
}
