/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


/* Input parser */

/* read in the grammar specification and record it in the format described in gram.h.
All actions are copied into faction,
in each case forming the body of a C function (yyaction)
which contains a switch statement to decide which action to execute.

The entry point is reader().  */

#include <fstream>
#include <sstream>
#include <string>

using namespace std;

#include <stdio.h>
#include <ctype.h>
#include "system.h"
#include "files.h"
#include "new.h"
#include "symtab.h"
#include "lex.h"
#include "gram.h"
#include "machine.h"

extern stringstream uniondeclaration;
extern stringstream headerdeclaration;
extern stringstream implementationdeclaration;

#ifdef _MSC_VER
 #pragma warning(disable : 4244)
#endif

extern bool nolinesflag;
extern bucket *symval;
extern int numval;
extern int failure;
extern int expected_conflicts;
extern char *token_buffer;

extern void init_lex();
extern void tabinit();
extern void output_headers();
extern void output_trailers();
extern void free_symtab();
extern void fatal(char *);
extern void fatals(char const*, char const*, char const*, char const*);
extern void unlex(int);
extern void done(int);

extern int skip_white_space();
extern int parse_percent_token();
extern int lex();

int read_signed_integer (istream & stream);
void read_declarations();
void copy_definition();
void parse_token_decl(int what_is, int what_is_not);
void parse_start_decl();
void parse_assoc_decl(int assoc);
void parse_type_decl();
void parse_union_decl(ostream & os);
void parse_expect_decl();
void copy_action();
void readgram();
void record_rule_line();
void packsymbols();

void EmitAboutToken(ostream &os);

void packgram();
int get_type();
string quoted_filename(string const&);
void copy_a_definition (ostream &);
void set_parser_name(string const&);

typedef struct symbol_list
{
	struct symbol_list *next;
	bucket *sym;
	bucket *ruleprec;
}
symbol_list;

// Number of slots allocated (but not necessarily used yet) in `rline'
int rline_allocated;

// Default name used when expanding parsernames in template-files
std::string parser_name="Parser"; 
bool parser_name_defined = false;

int lineno;
symbol_list *grammar;
int start_flag;
bucket *startval;
char **tags;

// Grammar definition globals, formerly in gram.cpp
int nitems;
int nrules;
int nsyms;
int ntokens;
int nvars;

short *ritem;
short *rlhs;
short *rrhs;
short *rprec;
short *rprecsym;
short *sprec;
short *rassoc;
short *sassoc;
short *token_translations;
short *rline;

int start_symbol;
int translations;
int max_user_token_number;
int error_token_number;
// End of grammar definition globals

// Nonzero if components of semantic values are used, implying they must be unions
static int value_components_used;
static int typed;  // nonzero if %union has been seen
static int lastprec;  // incremented for each %left, %right or %nonassoc seen
static int gensym_count;  // incremented for each generated symbol

bucket *errtoken;

extern char *version_string;

extern void output_before_read();
extern  void output_about_token();
void copy_header_definition();
void parse_name_declaration();
void parse_define();
void read_a_name();
void copy_action(symbol_list *rule, int stack_offset);

void reader()
{
	start_flag = 0;
	startval = NULL;  /* start symbol not specified yet. */
	
	// Translations is always set to 1, since we give `error' a user-token-number
	// to satisfy the Posix demand for YYERRCODE==256.
	translations = 1;
	
	nsyms = 1;
	nvars = 0;
	nrules = 0;
	nitems = 0;
	rline_allocated = 10;
	rline = NEW2(rline_allocated, short);
	
	typed = 0;
	lastprec = 0;
	
	gensym_count = 0;
	grammar = NULL;
	
	init_lex();
	lineno = 1;
	
	/* initialize the symbol table.  */
	tabinit();
	/* construct the error token */
	errtoken = getsym("error");
	errtoken->_class = STOKEN;
	errtoken->user_token_number = 256; /* Value specified by posix.  */
	/* construct a token that represents all undefined literal tokens. */
	/* it is always token number 2.  */
	getsym("$illegal.")->_class = STOKEN;
	/* Read the declaration section.  Copy %{ ... %} groups to ftable.
	Also notice any %token, %left, etc. found there.  */
	//output_before_read();
	
	read_declarations();
	/* read in the grammar, build grammar in list form.  write out actions.  */
	readgram();
	/* write closing delimiters for actions.  */
	//output_trailers();
	/* assign the symbols their symbol numbers. */
	packsymbols();
	/* convert the grammar into the format described in gram.h.  */
	packgram();
}


// read from finput until %% is seen.  Discard the %%.
// handle any % declarations, and copy the contents of any %{ ... %} groups to ftable.

void read_declarations ()
{
	int c;
	int tok;
	
	for (;;)
    {
		c = skip_white_space();
		
		if (c == '%')
		{
			tok = parse_percent_token();
			
			switch (tok)
			{
			case TWO_PERCENTS:
				return;
				
			case PERCENT_LEFT_CURLY:
				copy_definition();
				break;
			case PERCENT_LEFT_CURLY_HEADER:
				copy_header_definition();
				break;
				
			case TOKEN:
				parse_token_decl (STOKEN, SNTERM);
				break;
				
			case NTERM:
				parse_token_decl (SNTERM, STOKEN);
				break;
				
			case TYPE:
				parse_type_decl();
				break;
				
			case START:
				parse_start_decl();
				break;
				
			case UNION:
				parse_union_decl(uniondeclaration);
				break;
				
			case EXPECT:
				parse_expect_decl();
				break;
				
			case LEFT:
				parse_assoc_decl(LEFT_ASSOC);
				break;
				
			case RIGHT:
				parse_assoc_decl(RIGHT_ASSOC);
				break;
				
			case NONASSOC:
				parse_assoc_decl(NON_ASSOC);
				break;
				
			case PARSER_NAME:
				parse_name_declaration();
				break;
				
			case DEFINE_SYM:
				parse_define();
				break;
				
			default:
				fatal("junk after `%%' in definition section");
			}
		}
		else if (c == EOF)
			fatal("no input grammar");
		else if (c >= 040 && c <= 0177)
			fatals ("unknown character `%c' in declaration section", (char *)c, 0, 0);
		else
			fatals ("unknown character with code 0x%x in declaration section", (char *)c, 0, 0);
    }
}

/* copy the contents of a %{ ... %} into the definitions file.
The %{ has already been read.  Return after reading the %}.  */

void copy_definition()
{
	if (!nolinesflag)
		ftable << "#line " << lineno << " \"" << quoted_filename(infile) << "\"\n";
	copy_a_definition(implementationdeclaration);

}

void copy_header_definition()
{
	if (!nolinesflag)
		ftable << "#line " << lineno << " \"" << quoted_filename(infile) << "\"\n";
	copy_a_definition(headerdeclaration);
}

void copy_a_definition (ostream & os)
{
	bool ended;
	bool after_percent = false;  // true while reading a character if prev char was %
	int cplus_comment;
	
	int c = finput.get();
	
	for (;;)
    {
		switch (c)
		{
		case '\n':
			os.put(c);
			lineno++;
			break;
			
		case '%':
			after_percent = true;
			break;
			
		case '\'':
		case '"':
		{
			int const match = c;
			os.put(c);
			c = finput.get();
			
			while (c != match)
			{
				if (c == EOF || c == '\n')
					fatal("unterminated string");
				
				os.put(c);
				
				if (c == '\\')
				{
					c = finput.get();
					if (c == EOF)
						fatal("unterminated string");
					os.put(c);
					if (c == '\n')
						lineno++;
				}
				
				c = finput.get();
			}
			
			os.put(c);
			break;
		}	
		case '/':
			os.put(c);
			c = finput.get();
			if (c != '*' && c != '/')
				continue;
			
			cplus_comment = (c == '/');
			os.put(c);
			c = finput.get();
			
			ended = false;
			while (!ended)
			{
				if (!cplus_comment && c == '*')
				{
					while (c == '*')
					{
						os.put(c);
						c = finput.get();
					}
					
					if (c == '/')
					{
						os.put(c);
						ended = true;
					}
				}
				else if (c == '\n')
				{
					lineno++;
					os.put(c);
					if (cplus_comment)
						ended = true;
					else
						c = finput.get();
				}
				else if (c == EOF)
					fatal("unterminated comment in `%{' definition");
				else
				{
					os.put(c);
					c = finput.get();
				}
			}
			
			break;
			
		case EOF:
			fatal("unterminated `%{' definition");
			break;
			
		default:
			os.put(c);
		}
		
		c = finput.get();
		
		if (after_percent)
		{
			if (c == '}')
				return;
			os.put('%');

			after_percent = false;	
		}
    }
}


/* parse what comes after %token or %nterm.
For %token, what_is is STOKEN and what_is_not is SNTERM.
For %nterm, the arguments are reversed.  */

void parse_token_decl(int what_is, int what_is_not)
{
	int token = 0;
	int prev;
	char *type_name = 0;
	
	for (;;)
    {
		int c = skip_white_space();
		finput.unget();
		if (c == '%')
			return;
		
		// We have not passed a newline, so the token now starting is in this declaration
		prev = token;
		
		token = lex();
		if (token == COMMA)
			continue;
		if (token == TYPENAME)
		{
			type_name = strdup(token_buffer);
			value_components_used = 1;
		}
		else if (token == IDENTIFIER)
		{
			int oldclass = symval->_class;
			
			if (symval->_class == what_is_not)
				fatals("symbol %s redefined", (char *)(symval->tag), 0, 0);
			symval->_class = what_is;
			if (what_is == SNTERM && oldclass != SNTERM)
				symval->value = nvars++;
			
			if (type_name)
			{
				if (symval->type_name == NULL)
					symval->type_name = type_name;
				else
					fatals("type redeclaration for %s", (char *)(symval->tag), 0, 0);
			}
		}
		else if (prev == IDENTIFIER && token == NUMBER)
        {
			symval->user_token_number = numval;
			translations = 1;
        }
		else
			fatal("invalid text in %token or %nterm declaration");
    }	
}


// parse what comes after %start
void parse_start_decl ()
{
	if (start_flag)
		fatal("multiple %start declarations");
	start_flag = 1;
	if (lex() != IDENTIFIER)
		fatal("invalid %start declaration");
	startval = symval;
}


// read in a %type declaration and record its information for get_type_name to access
void parse_type_decl ()
{
	if (lex() != TYPENAME)
		fatal("ill-formed %type declaration");
	
	char * name = strdup(token_buffer);
	
	for (;;)
    {
		 int c = skip_white_space();
		 finput.unget();
		if(c == '%')
			return;
		
		// We have not passed a newline, so the token now starting is in this declaration
		
		int t = lex();
		switch (t)
		{
		case COMMA:
		case SEMICOLON:
			break;
		case IDENTIFIER:
			if (symval->type_name == NULL)
				symval->type_name = name;
			else
				fatals("type redeclaration for %s", (char *)(symval->tag), 0, 0);
			
			break;
		default:
			fatal("invalid %type declaration");
		}
    }
}


// read in a %left, %right or %nonassoc declaration and record its information.
// assoc is either LEFT_ASSOC, RIGHT_ASSOC or NON_ASSOC

void parse_assoc_decl (int assoc)
{
	char *name = NULL;
	int prev = 0;	// JF added = 0 to keep lint happy
	
	lastprec++;  // Assign a new precedence level, never 0.
	
	for (;;)
    {
		int t;
		
		int c = skip_white_space();
		finput.unget();
		if(c == '%')
			return;
		
		// We have not passed a newline, so the token now starting is in this declaration
		
		t = lex();
		
		switch (t)
		{
		case TYPENAME:
			name = strdup(token_buffer);
			break;
			
		case COMMA:
			break;
			
		case IDENTIFIER:
			if (symval->prec != 0)
				fatals("redefining precedence of %s", (char *)(symval->tag), 0, 0);
			symval->prec = lastprec;
			symval->assoc = assoc;
			if (symval->_class == SNTERM)
				fatals("symbol %s redefined", (char *)(symval->tag), 0, 0);
			symval->_class = STOKEN;
			if (name)
			{ /* record the type, if one is specified */
				if (symval->type_name == NULL)
					symval->type_name = name;
				else
					fatals("type redeclaration for %s", (char *)(symval->tag), 0, 0);
			}
			break;
			
		case NUMBER:
			if (prev == IDENTIFIER)
            {
				symval->user_token_number = numval;
				translations = 1;
            }
			else	  
				fatal("invalid text in association declaration");
			break;
			
		case SEMICOLON:
			return;
			
		default:
			fatal("malformatted association declaration");
		}
		
		prev = t;
		
    }
}


// copy the union declaration into ftable, where it is made into the
// definition of YYSTYPE, the type of elements of the parser value stack.

void parse_union_decl(ostream & os)
{
	int c;
	int count;
	int in_comment;
	int cplus_comment;
	
	if (typed)
		fatal("multiple %union declarations");
	
	typed = 1;
	
	if (!nolinesflag)
		ftable << "#line " << lineno << " \"" << quoted_filename(infile) << "\"\n";

	count = 0;
	in_comment = 0;
	
	c = finput.get();
	
	while (c != EOF)
    {
		switch (c)
		{
		case '\n':
			if (count > 0)
				os << '\n';
			lineno++;
			break;
			
		case '/':
			os << (char)c;
			c = finput.get();
			if (c != '*' && c != '/')
				finput.unget();
			else
			{
				os << (char)c;
				cplus_comment = (c == '/');
				in_comment = 1;
				c = finput.get();
				while (in_comment)
				{
					os << (char)c;
					if (c == '\n')
					{
						lineno++;
						if (cplus_comment)
						{
							in_comment = 0;
							break;
						}
					}
					if (c == EOF)
						fatal("unterminated comment");
					
					if (!cplus_comment && c == '*')
					{
						c = finput.get();
						if (c == '/')
						{
							os << '/';
							in_comment = 0;
						}
					}
					else
						c = finput.get();
				}
			}
			break;
			
			
		case '{':
			if (count > 0)
				os << '{';
			count++;
			break;
			
		case '}':
			if (count == 0)
				fatal ("unmatched close-brace (`}')");
			count--;
			if (count > 0)
				os << '}';
			else
			{
				c=skip_white_space();
				if(c!=';') finput.unget();
				return;
			}
			break;
		default:
			os << (char)c;
			break;
		}
		
		c = finput.get();
    }
}

/* parse the declaration %expect N which says to expect N
shift-reduce conflicts.  */

void parse_expect_decl()
{
	int c;
	int count;
	char buffer[20];
	
	c = finput.get();
	while (c == ' ' || c == '\t')
		c = finput.get();
	
	count = 0;
	while (c >= '0' && c <= '9')
    {
		if (count < 20)
			buffer[count++] = c;
		c = finput.get();
    }
	buffer[count] = 0;
	
	finput.unget();
	expected_conflicts = atoi(buffer);
}

/* that's all of parsing the declaration section */

/* Get the data type (alternative in the union) of the value for symbol n in rule rule.  */

char *get_type_name(int n, symbol_list *rule)
{
	static char *msg = "invalid $ value";
	
	int i;
	symbol_list *rp;
	
	if (n < 0)
		fatal(msg);
	
	rp = rule;
	i = 0;
	
	while (i < n)
    {
		rp = rp->next;
		if (rp == NULL || rp->sym == NULL)
			fatal(msg);
		i++;
    }
	
	return (rp->sym->type_name);
}

/* Assuming that a { has just been seen, copy everything up to the matching }
into the actions file.
stack_offset is the number of values in the current rule so far,
which says where to find $0 with respect to the top of the stack.  */

void copy_action(symbol_list *rule, int stack_offset)
{
	int n;
	int count;
	bool ended;
	char *type_name;
	int cplus_comment;
	
	faction << "\ncase " << nrules << ":\n";
	if (!nolinesflag)
		faction << "#line " << lineno << " \"" << quoted_filename(infile) << "\"\n";
	faction.put('{');
	
	count = 1;
	int c = finput.get();
	
	while (count > 0)
    {
		while (c != '}')
        {
			switch (c)
			{
			case '\n':
				faction.put(c);
				lineno++;
				break;
				
			case '{':
				faction.put(c);
				count++;
				break;
				
			case '\'':
			case '"':
			{
				int const match = c;
				faction.put(c);
				c = finput.get();
				
				while (c != match)
				{
					if (c == EOF || c == '\n')
						fatal("unterminated string");
					
					faction.put(c);
					
					if (c == '\\')
					{
						c = finput.get();
						if (c == EOF)
							fatal("unterminated string");
						faction.put(c);
						if (c == '\n')
							lineno++;
					}
					
					c = finput.get();
				}
				
				faction.put(c);
				break;
			}
				
			case '/':
				faction.put(c);
				c = finput.get();
				if (c != '*' && c != '/')
					continue;
				
				cplus_comment = (c == '/');
				faction.put(c);
				c = finput.get();
				
				ended = false;
				while (!ended)
				{
					if (!cplus_comment && c == '*')
					{
						while (c == '*')
						{
							faction.put(c);
							c = finput.get();
						}
						
						if (c == '/')
						{
							faction.put(c);
							ended = true;
						}
					}
					else if (c == '\n')
					{
						lineno++;
						faction.put(c);
						if (cplus_comment)
							ended = true;
						else
							c = finput.get();
					}
					else if (c == EOF)
						fatal("unterminated comment");
					else
					{
						faction.put(c);
						c = finput.get();
					}
				}
				
				break;
				
			case '$':
				c = finput.get();
				type_name = NULL;
				
				if (c == '<')
				{
					char *cp = token_buffer;
					
					while ((c = finput.get()) != '>' && c > 0)
						*cp++ = c;
					*cp = 0;
					type_name = token_buffer;
					value_components_used = 1;
					
					c = finput.get();
				}
				if (c == '$')
				{
					faction << "yyval";
					if (!type_name) type_name = get_type_name(0, rule);
					if (type_name)
						faction << '.' << type_name;
					if(!type_name && typed)	/* JF */
						fprintf(stderr,"%s(%d) : warning: $$ of '%s' has no declared type.\n",infile.c_str(),lineno,rule->sym->tag);
				}
				else if (isdigit(c) || c == '-')
				{
					finput.unget();
					n = read_signed_integer(finput);
					c = finput.get();
					
					if (!type_name && n > 0)
						type_name = get_type_name(n, rule);
					
					faction << "yyvsp[" << n - stack_offset << ']';
					if (type_name)
						faction << '.' << type_name;
					if(!type_name && typed)	/* JF */
						fprintf(stderr,"%s(%d) : warning: $%d of '%s' has no declared type.\n",infile.c_str(),lineno,n,rule->sym->tag);
					continue;
				}
				else
					fatals("$%c is invalid", (char *)c, 0, 0);
				
				break;
				
			case '@':
				c = finput.get();
				if (isdigit(c) || c == '-')
				{
					finput.unget();
					n = read_signed_integer(finput);
					c = finput.get();
				}
				else
				{
					fatal("invalid @-construct");
					n = 0;
				}
				
				faction << "yylsp[" << n - stack_offset << ']';
				continue;
				
			case EOF:
				fatal("unmatched '{'");
				
			default:
				faction.put(c);
		}
		
		c = finput.get();
        }
		
		// above loop exits when c is '}'
		if (--count)
        {
			faction.put(c);
			c = finput.get();
		}
    }
	
	faction << ";\n\tbreak; }";
}


// Generate a dummy symbol, a nonterminal, whose name cannot conflict
// with the user's names

bucket *gensym()
{
	bucket *sym;

	sprintf (token_buffer, "@%d", ++gensym_count);
	sym = getsym(token_buffer);
	sym->_class = SNTERM;
	sym->value = nvars++;
	return (sym);
}

/* Parse the input grammar into a one symbol_list structure.
Each rule is represented by a sequence of symbols: the left hand side
followed by the contents of the right hand side, followed by a null pointer
instead of a symbol to terminate the rule.
The next symbol is the lhs of the following rule.

All actions are copied out to the appropriate files,
labelled by the rule number they apply to.  */

void readgram()
{
	int t;
	bucket *lhs = 0;
	symbol_list *p;
	symbol_list *p1;
	bucket *bp;
	
	symbol_list *crule;	/* points to first symbol_list of current rule.  */
	/* its symbol is the lhs of the rule.   */
	symbol_list *crule1;  /* points to the symbol_list preceding crule.  */
	
	p1 = NULL;
	
	t = lex();
	
	while (t != TWO_PERCENTS && t != ENDFILE)
	{
		if (t == IDENTIFIER || t == BAR)
		{
			int actionflag = 0;
			int rulelength = 0;  /* number of symbols in rhs of this rule so far  */
			int xactions = 0;	/* JF for error checking */
			bucket *first_rhs = 0;

			if (t == IDENTIFIER)
			{
				lhs = symval;

				t = lex();
				if (t != COLON)
					fatal("ill-formed rule");
			}

			if (nrules == 0)
			{
				if (t == BAR)
					fatal("grammar starts with vertical bar");

				if (!start_flag)
					startval = lhs;
			}

			/* start a new rule and record its lhs.  */

			nrules++;
			nitems++;

			record_rule_line ();

			p = NEW(symbol_list);
			p->sym = lhs;

			crule1 = p1;
			if (p1)
				p1->next = p;
			else
				grammar = p;

			p1 = p;
			crule = p;

			/* mark the rule's lhs as a nonterminal if not already so.  */

			if (lhs->_class == SUNKNOWN)
			{
				lhs->_class = SNTERM;
				lhs->value = nvars;
				nvars++;
			}
			else if (lhs->_class == STOKEN)
				fatals("rule given for %s, which is a token", (char *)lhs->tag, 0, 0);

			/* read the rhs of the rule.  */

			for (;;)
			{
				t = lex();

				if (! (t == IDENTIFIER || t == LEFT_CURLY)) break;

				/* If next token is an identifier, see if a colon follows it.
				If one does, exit this rule now.  */
				if (t == IDENTIFIER)
				{
					bucket *ssave;
					int t1;

					ssave = symval;
					t1 = lex();
					unlex(t1);
					symval = ssave;
					if (t1 == COLON) break;

					if(!first_rhs)	/* JF */
						first_rhs = symval;
					/* Not followed by colon =>
					process as part of this rule's rhs.  */
				}

				/* If we just passed an action, that action was in the middle
				of a rule, so make a dummy rule to reduce it to a
				non-terminal.  */
				if (actionflag)
				{
					bucket *sdummy;

					/* Since the action was written out with this rule's */
					/* number, we must write give the new rule this number */
					/* by inserting the new rule before it.  */

					/* Make a dummy nonterminal, a gensym.  */
					sdummy = gensym();

					/* Make a new rule, whose body is empty,
					before the current one, so that the action
					just read can belong to it.  */
					nrules++;
					nitems++;
					record_rule_line ();
					p = NEW(symbol_list);
					if (crule1)
						crule1->next = p;
					else grammar = p;
					p->sym = sdummy;
					crule1 = NEW(symbol_list);
					p->next = crule1;
					crule1->next = crule;

					/* insert the dummy generated by that rule into this rule.  */
					nitems++;
					p = NEW(symbol_list);
					p->sym = sdummy;
					p1->next = p;
					p1 = p;

					actionflag = 0;
				}

				if (t == IDENTIFIER)
				{
					nitems++;
					p = NEW(symbol_list);
					p->sym = symval;
					p1->next = p;
					p1 = p;
				}
				else /* handle an action.  */
				{
					copy_action(crule, rulelength);
					actionflag = 1;
					xactions++;	/* JF */
				}
				rulelength++;
			}

			/* Put an empty link in the list to mark the end of this rule  */
			p = NEW(symbol_list);
			p1->next = p;
			p1 = p;

			if (t == PREC)
			{
				t = lex();
				crule->ruleprec = symval;
				t = lex();
			}
			if (t == LEFT_CURLY)
			{
				if (actionflag) fatal("two actions at end of one rule");
				copy_action(crule, rulelength);
				t = lex();
			}
			/* If $$ is being set in default way,
			warn if any type mismatch.  */
			else if (!xactions && first_rhs && lhs->type_name != first_rhs->type_name)
			{
				if (lhs->type_name == 0 || first_rhs->type_name == 0
					|| strcmp(lhs->type_name,first_rhs->type_name))
					fprintf(stderr, "%s(%d) : warning: type clash ('%s' '%s') on default action\n",
					infile.c_str(),
					lineno,
					lhs->type_name ? lhs->type_name : "",
					first_rhs->type_name ? first_rhs->type_name : "");
			}
			/* Warn if there is no default for $$ but we need one.  */
			else if (!xactions && !first_rhs && lhs->type_name != 0)
				fprintf(stderr,
				"%s(%d) : warning: empty rule for typed nonterminal, and no action\n",
				infile.c_str(),
				lineno);
			if (t == SEMICOLON)
				t = lex();
		}
		/* these things can appear as alternatives to rules.  */
		else switch(t)
		{
			case TOKEN:
				parse_token_decl(STOKEN, SNTERM);
				t = lex();
				break;
			case NTERM:
				parse_token_decl(SNTERM, STOKEN);
				t = lex();
				break;
			case TYPE:
				t = get_type();
				break;
			case UNION:
				parse_union_decl(uniondeclaration);
				t = lex();
				break;
			case EXPECT:
				parse_expect_decl();
				t = lex();
			break;
			case START:
				parse_start_decl();
				t = lex();
			break;	
			default:
				fatal("invalid input");
				break;
		}
	}
	
	if (nsyms > MAXSHORT)
		fatals("too many symbols (tokens plus nonterminals); maximum %d", (char *)MAXSHORT, 0, 0);
	if (nrules == 0)
		fatal("no input grammar");
	
	/* Report any undefined symbols and consider them nonterminals.  */
	
	for (bp = firstsymbol; bp; bp = bp->next)
	{
		if (bp->_class == SUNKNOWN)
		{
			fprintf(stderr, "symbol %s used, not defined as token, and no rules for it\n",
				bp->tag);
			failure = 1;
			bp->_class = SNTERM;
			bp->value = nvars++;
		}
	}
	
	ntokens = nsyms - nvars;
}


void record_rule_line ()
{
	// Record each rule's source line number in rline table
	if (nrules >= rline_allocated)
    {
		rline_allocated = nrules * 2;
		rline = (short *) realloc (rline, rline_allocated * sizeof (short));
    }
	rline[nrules] = lineno;
}


// read in a %type declaration and record its information for get_type_name to access
int get_type()
{
	int t = lex();
	
	if (t != TYPENAME)
		fatal("ill-formed %type declaration");
	
	char * name = strdup(token_buffer);
	for (;;)
    {
		t = lex();
		
		switch (t)
		{
		case SEMICOLON:
			return (lex());
			
		case COMMA:
			break;
			
		case IDENTIFIER:
			if (symval->type_name == NULL)
				symval->type_name = name;
			else
				fatals("type redeclaration for %s", (char *)symval->tag, 0, 0);
			
			break;
			
		default:
			return (t);
		}
    }
}


/* assign symbol numbers, and write definition of token names into fdefines.
Set up vectors tags and sprec of names and precedences of symbols.  */

void packsymbols()
{
	bucket *bp;
	int tokno = 1;
	
	tags = NEW2(nsyms + 1, char *);
	tags[0] = "$";
	
	sprec = NEW2(nsyms, short);
	sassoc = NEW2(nsyms, short);
	
	max_user_token_number = 256;
	int last_user_token_number = 256;
	
	for (bp = firstsymbol; bp; bp = bp->next)
    {
		if (bp->_class == SNTERM)
			bp->value += ntokens;
		else
		{
			if (translations && !(bp->user_token_number))
				bp->user_token_number = ++last_user_token_number;
			if (bp->user_token_number > max_user_token_number)
				max_user_token_number = bp->user_token_number;
			bp->value = tokno++;
		}
		
		tags[bp->value] = bp->tag;
		sprec[bp->value] = bp->prec;
		sassoc[bp->value] = bp->assoc;		
    }
	
	if (translations)
    {
		token_translations = NEW2(max_user_token_number+1, short);
		
		/* initialize all entries for literal tokens to 2,
		the internal token number for $illegal., which represents all invalid inputs.  */
		for (int i = 0; i <= max_user_token_number; i++)
			token_translations[i] = 2;      
    }
	
	for (bp = firstsymbol; bp; bp = bp->next)
    {
		if (bp->value >= ntokens) continue;
		if (translations)
		{
			if (token_translations[bp->user_token_number] != 2)
			{
				/* JF made this a call to fatals() */
				fatals( "tokens %s and %s both assigned number %d",
					(char *)tags[token_translations[bp->user_token_number]],
					(char *)bp->tag,
					(char *)(int)bp->user_token_number);
			}
			token_translations[bp->user_token_number] = bp->value;
		}
    }
	
	error_token_number = errtoken->value;
		
	if (startval->_class == SUNKNOWN)
		fatals("the start symbol %s is undefined", (char *)startval->tag, 0, 0);
	else if (startval->_class == STOKEN)
		fatals("the start symbol %s is a token", (char *)startval->tag, 0, 0);
	
	start_symbol = startval->value;
}


/* convert the rules into the representation using rrhs, rlhs and ritems.  */

void packgram()
{	
	ritem = NEW2(nitems + 1, short);
	rlhs = NEW2(nrules, short) - 1;
	rrhs = NEW2(nrules, short) - 1;
	rprec = NEW2(nrules, short) - 1;
	rprecsym = NEW2(nrules, short) - 1;
	rassoc = NEW2(nrules, short) - 1;
	
	int itemno = 0;
	int ruleno = 1;

	symbol_list *p = grammar;
	while (p)
    {
		rlhs[ruleno] = p->sym->value;
		rrhs[ruleno] = itemno;
		bucket *ruleprec = p->ruleprec;
		
		p = p->next;
		while (p && p->sym)
		{
			ritem[itemno++] = p->sym->value;
			/* A rule gets by default the precedence and associativity
			of the last token in it.  */
			if (p->sym->_class == STOKEN)
			{
				rprec[ruleno] = p->sym->prec;
				rassoc[ruleno] = p->sym->assoc;
			}
			if (p) p = p->next;
		}
		
		/* If this rule has a %prec,
		the specified symbol's precedence replaces the default.  */
		if (ruleprec)
		{
			rprec[ruleno] = ruleprec->prec;
			rassoc[ruleno] = ruleprec->assoc;
			rprecsym[ruleno] = ruleprec->value;
		}
		
		ritem[itemno++] = -ruleno;
		ruleno++;
		
		if (p) p = p->next;
    }
	
	ritem[itemno] = 0;
}

int read_signed_integer(istream & stream)
{
	int i;
	stream >> i;
	return i;
}

void set_parser_name(string const& name)
{
	if(parser_name_defined)
		fatals("parser name already defined as \"%s\"and redefined to \"%s\"\n", parser_name.c_str(), name.c_str(), 0);

	parser_name = name;
	parser_name_defined = true;
}

void read_a_name(char *buf, int len)
{
	int c,l;
	for(c = skip_white_space(),l=0; (isalnum(c) || c == '_'); c=finput.get(),l++) 
		if(l<len) buf[l]=c;

	if(l>=len)
	{
		buf[len-1]=0;
		fprintf(stderr,"%s:%d name too long, truncated to :\"%s\"\n"
			,infile.c_str(),lineno,buf);
	}
	else
		buf[l]=0;    

	finput.unget();
}

void parse_name_declaration()
{
	char name[65];
	read_a_name(name,sizeof(name));
	set_parser_name(name);
}

void parse_define()
{
	char name[65];
	int c;
	int after_backslash;
	read_a_name(name, sizeof(name));
	ftable << "#define YY_" << parser_name << '_' << name << ' ';
	for(after_backslash=0, c=finput.get(); (after_backslash || c!='\n'); c=finput.get())
	{
		after_backslash=(c=='\\');
		if(c=='\n') lineno++;
		if(c==EOF)
		{fatal("unexpected EOF in %define");}
		ftable.put(c);
	}
	ftable.put('\n');
	lineno++;
}

string quoted_filename(string const& f)
{
	stringstream ss;

	string::const_iterator iter = f.begin();
	for(; iter != f.end(); iter++)
	{
		if (*iter == '\\')
			ss << '\\';
		ss << *iter;
	}
	return ss.str();
}
