/*
Mammut parser generator
Copyright (C) 2005 Lumai HB - http://www.lumai.se
This source code is based on Bison++ by Alain Coetmeur.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
02110-1301, USA
*/


/* Grammar reduction */

/* Reduce the grammar:  Find and eliminate unreachable terminals,
nonterminals, and productions.  David S. Bakin.

Don't eliminate unreachable terminals:  They may be used by the user's parser. */

#include <stdio.h>
#include "system.h"
#include "files.h"
#include "gram.h"
#include "machine.h"
#include "new.h"

#include <iostream>
#include <iomanip>

using namespace std;

extern char **tags;		// reader.c
extern bool verboseflag;

typedef unsigned *BSet;
typedef short  *rule;

/*	N is set of all nonterminals which are not useless.  P is set of all rules
	which have no useless nonterminals in their RHS.  V is the set of all
	accessible symbols. */

static BSet N, P, V, V1;

static int nuseful_productions, nuseless_productions, nuseful_nonterminals,
	nuseless_nonterminals;

static void useless_nonterminals();
static void inaccessable_symbols();
static void reduce_grammar_tables();
static void print_results();
static void print_notices();
void dump_grammar();

extern void fatals (char const*, char const*, char const*, char const*);

bool bits_equal(BSet L, BSet R, int n)
{
	for (int i = n - 1; i >= 0; i--)
		if (L[i] != R[i])
			return false;
	return true;
}

// Return number of 1 bits in an array of n bitsets
int bits_size(BSet S, int n)
{
	int count = 0;	
	for (int i = n-1; i >= 0; i--)
	{
		unsigned curBit = S[i];
		while (curBit)
		{
			count++;
			curBit &= (curBit-1);
		}
	}
	return count;
}

void reduce_grammar ()
{
	bool reduced;
	
	/* Allocate the global sets used to compute the reduced grammar */
	
	N = NEW2(WORDSIZE(nvars), unsigned);
	P = NEW2(WORDSIZE(nrules + 1), unsigned);
	V = NEW2(WORDSIZE(nsyms), unsigned);
	V1 = NEW2(WORDSIZE(nsyms), unsigned);
	
	useless_nonterminals();
	inaccessable_symbols();
	
	reduced = (bool) (nuseless_nonterminals + nuseless_productions > 0);
	
	if (verboseflag)
		print_results();
	
	if (reduced)
	{
		print_notices();

		if (!BITISSET(N, start_symbol - ntokens))
			fatals("Start symbol %s does not derive any sentence.", (char *)tags[start_symbol], 0, 0);

		reduce_grammar_tables();
	}

	// Free the global sets used to compute the reduced grammar */	
	FREE(N);
	FREE(V);
	FREE(P);
}

/*
* Another way to do this would be with a set for each production and then do
* subset tests against N, but even for the C grammar the whole reducing
* process takes only 2 seconds on my 8Mhz AT.
*/

static bool useful_production (int  i, BSet N)
{
	/* A production is useful if all of the nonterminals in its RHS
	* appear in the set of useful nonterminals */
	
	for (rule r = &ritem[rrhs[i]]; *r > 0; r++)
	{
		short const n = *r;
		if (ISVAR(n))
			if (!BITISSET(N, n - ntokens))
				return false;
	}
	return true;
}

/* Remember that rules are 1-origin, symbols are 0-origin. */

static void useless_nonterminals ()
{
	BSet Np, Ns;
	int  i, n;
	
	/*
	* N is set as built.  Np is set being built this iteration. P is set
	* of all productions which have a RHS all in N.
	*/
	
	Np = NEW2(WORDSIZE(nvars), unsigned);
	
	/*
	* The set being computed is a set of nonterminals which can derive
	* the empty string or strings consisting of all terminals. At each
	* iteration a nonterminal is added to the set if there is a
	* production with that nonterminal as its LHS for which all the
	* nonterminals in its RHS are already in the set.  Iterate until the
	* set being computed remains unchanged.  Any nonterminals not in the
	* set at that point are useless in that they will never be used in
	* deriving a sentence of the language.
	* 
	* This iteration doesn't use any special traversal over the
	* productions.  A set is kept of all productions for which all the
	* nonterminals in the RHS are in useful.  Only productions not in
	* this set are scanned on each iteration.  At the end, this set is
	* saved to be used when finding useful productions: only productions
	* in this set will appear in the final grammar.
	*/
	
	n = 0;
	while (true)
    {
		for (i = WORDSIZE(nvars) - 1; i >= 0; i--)
			Np[i] = N[i];
		for (i = 1; i <= nrules; i++)
		{
			if (!BITISSET(P, i))
			{
				if (useful_production(i, N))
				{
					SETBIT(Np, rlhs[i] - ntokens);
					SETBIT(P, i);
				}
			}
		}
		if (bits_equal(N, Np, WORDSIZE(nvars)))
			break;
		Ns = Np;
		Np = N;
		N = Ns;
    }
	FREE(N);
	N = Np;
}

static void inaccessable_symbols ()
{
	BSet  Vp, Vs, Pp;
	int   i, n;
	short t;
	rule  r;
	
	/*
	* Find out which productions are reachable and which symbols are
	* used.  Starting with an empty set of productions and a set of
	* symbols which only has the start symbol in it, iterate over all
	* productions until the set of productions remains unchanged for an
	* iteration.  For each production which has a LHS in the set of
	* reachable symbols, add the production to the set of reachable
	* productions, and add all of the nonterminals in the RHS of the
	* production to the set of reachable symbols.
	* 
	* Consider only the (partially) reduced grammar which has only
	* nonterminals in N and productions in P.
	* 
	* The result is the set P of productions in the reduced grammar, and
	* the set V of symbols in the reduced grammar.
	* 
	* Although this algorithm also computes the set of terminals which are
	* reachable, no terminal will be deleted from the grammar. Some
	* terminals might not be in the grammar but might be generated by
	* semantic routines, and so the user might want them available with
	* specified numbers.  (Is this true?)  However, the nonreachable
	* terminals are printed (if running in verbose mode) so that the user
	* can know.
	*/
	
	Vp = NEW2(WORDSIZE(nsyms), unsigned);
	Pp = NEW2(WORDSIZE(nrules + 1), unsigned);
	
	/* If the start symbol isn't useful, then nothing will be useful. */
	if(BITISSET(N, start_symbol - ntokens))
	{	
		SETBIT(V, start_symbol);

		n = 0;
		while (true)
		{
			for (i = WORDSIZE(nsyms) - 1; i >= 0; i--)
				Vp[i] = V[i];

			for (i = 1; i <= nrules; i++)
			{
				if (!BITISSET(Pp, i) && BITISSET(P, i) && BITISSET(V, rlhs[i]))
				{
					for (r = &ritem[rrhs[i]]; *r >= 0; r++)
					{
						if (ISTOKEN(t = *r) || BITISSET(N, t - ntokens))
							SETBIT(Vp, t);
					}
					SETBIT(Pp, i);
				}
			}

			if (bits_equal(V, Vp, WORDSIZE(nsyms)))
				break;

			Vs = Vp;
			Vp = V;
			V = Vs;
		}
	}
	
	FREE(V);
	V = Vp;
	
	// Tokens 0, 1, and 2 are internal to Bison.  Consider them useful.
	SETBIT(V, 0);	// end-of-input token
	SETBIT(V, 1);	// error token
	SETBIT(V, 2);	//illegal token
	
	FREE(P);
	P = Pp;
	
	nuseful_productions = bits_size(P, WORDSIZE(nrules + 1));
	nuseless_productions = nrules - nuseful_productions;
	
	nuseful_nonterminals = 0;
	for (i = ntokens; i < nsyms; i++)
		if (BITISSET(V, i))
			nuseful_nonterminals++;

	nuseless_nonterminals = nvars - nuseful_nonterminals;

	// A token that was used in %prec should not be warned about.
	for (i = 1; i < nrules; i++)
		if (rprecsym[i] != 0)
			SETBIT(V1, rprecsym[i]);
}

static void reduce_grammar_tables ()
{
	// Disable useless productions, since they may contain useless nonterms
	// that would get mapped below to -1 and confuse everyone.
	if (nuseless_productions > 0)
    {
		for (int pn = 1; pn <= nrules; pn++)
		{
			if (!BITISSET(P, pn))
				rlhs[pn] = -1;
		}
    }
	
	/* remove useless symbols */
	if (nuseless_nonterminals > 0)
    {		
		//  Create a map of nonterminal number to new nonterminal
		//	number. -1 in the map means it was useless and is being eliminated.

		short *nontermmap = NEW2(nvars, short) - ntokens;
		int i;
		for (i = ntokens; i < nsyms; i++)
			nontermmap[i] = -1;

		int n = ntokens;
		for (i = ntokens; i < nsyms; i++)
			if (BITISSET(V, i))
				nontermmap[i] = short(n++);

		// Shuffle elements of tables indexed by symbol number.
		for (i = ntokens; i < nsyms; i++)
		{
			n = nontermmap[i];
			if (n >= 0)
			{
				sassoc[n] = sassoc[i];
				sprec[n] = sprec[i];
				tags[n] = tags[i];
			}
			else
				free(tags[i]);
		}

		// Replace all symbol numbers in valid data structures.
		for (i = 1; i <= nrules; i++)
		{
			// Ignore the rules disabled above.
			if (rlhs[i] >= 0)
				rlhs[i] = nontermmap[rlhs[i]];
			if (ISVAR (rprecsym[i])) // Can this happen?
				rprecsym[i] = nontermmap[rprecsym[i]];
		}

		for (rule r = ritem; *r; r++)
			if (ISVAR(*r))
				*r = nontermmap[*r];

		start_symbol = nontermmap[start_symbol];

		nsyms -= nuseless_nonterminals;
		nvars -= nuseless_nonterminals;

		free(&nontermmap[ntokens]);
	}
}

static void print_results ()
{
	rule  r;
	bool  b;
	
	if (nuseless_nonterminals > 0)
    {
		clog << "Useless nonterminals:\n\n";
		for (int i = ntokens; i < nsyms; i++)
			if (!BITISSET(V, i))
				clog << "   " << tags[i] << '\n';
    }
	b = false;
	for (int i = 0; i < ntokens; i++)
    {
		if (!BITISSET(V, i) && !BITISSET(V1, i))
		{
			if (!b)
			{
				clog << "\n\nTerminals which are not used:\n\n";
				b = true;
			}
			clog << "   " << tags[i] << '\n';
		}
    }
	
	if (nuseless_productions > 0)
    {
		clog << "\n\nUseless rules:\n\n";
		for (int i = 1; i <= nrules; i++)
		{
			if (!BITISSET(P, i))
			{
				clog << setw(4) << i << "  ";
				clog << tags[rlhs[i]] << " :\t";
				for (r = &ritem[rrhs[i]]; *r >= 0; r++)
					clog << ' ' << tags[*r];
				clog << ";\n";
			}
		}
    }
	if (nuseless_nonterminals > 0 || nuseless_productions > 0 || b)
		clog << "\n\n";
}

void dump_grammar ()
{
	int i;
	rule r;
	
	clog << "ntokens = " << ntokens << ", nvars = " << nvars << ", nsyms = " << nsyms << ", nrules = " << nrules << ", nitems = " << nitems << "\n\n";
	clog << "Variables\n---------\n\n";
	clog << "Value  Sprec    Sassoc    Tag\n";
	for (i = ntokens; i < nsyms; i++)
		clog << setw(7) << i << sprec[i] << sassoc[i] << tags[i] << '\n';
	clog << "\n\n";
	clog << "Rules\n-----\n\n";
	for (i = 1; i <= nrules; i++)
    {
		clog << setw(5) << i << '(' << setw(5) << rprec[i] << setw(5) << rassoc[i] << ')' << rlhs[i] << " : (@" << rrhs[i] << ')';
		for (r = &ritem[rrhs[i]]; *r > 0; r++)
			clog << setw(5) << *r;
		clog << " [" << -(*r) << "]\n";
    }
	clog << "\n\n";
	clog << "Rules interpreted\n-----------------\n\n";
	for (i = 1; i <= nrules; i++)
    {
		clog << setw(5) << i << "  " << tags[rlhs[i]];
		for (r = &ritem[rrhs[i]]; *r > 0; r++)
			clog << ' ' << tags[*r];
		clog << '\n';
    }
	clog << "\n\n";
}

static void print_notices ()
{
	fprintf(stderr, "%s contains ", infile.c_str());
	
	if (nuseless_nonterminals > 0)
    {
		fprintf(stderr, "%d useless nonterminal%s",
			nuseless_nonterminals, (nuseless_nonterminals == 1 ? "" : "s"));
    }
	if (nuseless_nonterminals > 0 && nuseless_productions > 0)
		fprintf(stderr, " and ");
	
	if (nuseless_productions > 0)
    {
		fprintf(stderr, "%d useless rule%s",
			nuseless_productions, (nuseless_productions == 1 ? "" : "s"));
    }
	fprintf(stderr, ".\n");
	fflush(stderr);
}
